/*

 unfold_c

 Join adjacent lines in a text file, to undo the effect of a FOLD

 Date		Author		Comments
 ----		------		--------
 14/03/1998	A. D. Ives	Original.

 1.01
 ----
 15/03/1998			Constants added for program version and
				block size to be used when reading stdin.
				New -s option to skip the first n lines.

 1.02
 ----
 22/03/1998			Added support for a list of characters
				which indicate that the following line
				should not be joined to the current one.
				Available through -n switch.
 23/02/1998			Added code to trap a trailing blank at the
				end of a line so no longer creates a double
				space.
				Fixed bug that did not join a line if the
				chr10 was the very last byte in the buffer
				New -S option to skip until after a line has
				been found that begins with the supplied
				string.

 1.03
 ----
 25/03/1998			Added support for -Nt option.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define TRUE 1
#define FALSE 0
#define BLOCK_SIZE 4096

char _PROG_NAME[]= "unfold" ;
#define PROGRAM_VERSION "1.03"

char NoJoin[32], SkipTo[128] ;

void help() {

  /* Display the options */

  printf("\nUnfold v%s (c) Adrian D. Ives\n\n", PROGRAM_VERSION) ;

  printf("Join together adjacent lines on stdin to undo the effects\n") ;
  printf("of a previous FOLD, and remove any Carriage Returns from\n") ;
  printf("the input stream. Send the result to stdout.\n\n") ;

  printf("Switches:-\n\n") ;
  printf("-k    Keep Carriage Returns\n") ;
  printf("-r    Just remove Carriage Returns (no joins)\n") ;
  printf("-sN   Skip (do not join) the first N lines\n") ;
  printf("-SC   Skip (do not join) until after line beginning with C\n") ;
  printf("-nC   Do not join lines beginning with any of chars in C\n") ;
  printf("-N    Do not join lines beginning with a space\n") ;
  printf("-Nt   Do not join lines beginning with a Tab\n\n") ;

}

char *strstripch(char *str1, char c) {

  /* Strip character c from string str1 */

  char *ptr1, *ptr2 ;

  if ( c ) {
    ptr1= ptr2= str1 ;
    while( *ptr1 ) {
      if ( *ptr1 != c )
	*ptr2++ = *ptr1 ;
      ptr1++ ;
    }
    *ptr2= 0 ;
  }
  return str1 ;
}

main(int argc, char **argv) {

  char buff[BLOCK_SIZE+2], *ptr, next ;
  int i,
      bytes,
      StripCR,
      JoinLines,
      SkipCount,
      SkipToLen ;

  StripCR= JoinLines= TRUE ;
  SkipCount= 0 ;
  *NoJoin= *SkipTo= 0 ;

  for ( i= 1 ; i <= argc - 1 ; i++ ) {
    if ( argv[i][0] == '-' ) {
      switch( argv[i][1] ) {
	case  0  :
	case '?' :
	  help() ;
	  return 0 ;
	case 'k' :
	  StripCR= FALSE ;
	  break ;
	case 'n' :
	  strcpy(NoJoin, argv[i] + 2) ;
	  break ;
	case 'N' :
	  if ( argv[i][2] == 't' || argv[i][2] == 'T' )
	    strcat(NoJoin, "\t") ;
	  else
	    strcat(NoJoin, " ") ;
	  break ;
	case 'r' :
	  JoinLines= FALSE ;
	  break ;
	case 's' :
	  if ( strlen(argv[i]) > 2 )
	    SkipCount= atoi(argv[i]+2) ;
	  break ;
	case 'S' :
	  strcpy(SkipTo, argv[i] + 2) ;
	  break ;
      }
    }
  }

  SkipToLen= strlen(SkipTo) ;
  bytes= 1 ;
  while ( bytes && !feof(stdin) ) {
    bytes= fread(buff, 1, BLOCK_SIZE, stdin) ;
    if ( bytes > 0 ) {
      buff[bytes]= '\0' ;
      if ( StripCR )
	strstripch(buff, 13) ;
      if ( JoinLines ) {
	ptr= buff ;
	while ( (i= strpos(ptr, '\n')) != -1 ) {
	  ptr+= i ;
	  if ( !*(ptr + 1) ) {          /* Hit end of buffer            */
	    if ( fread(ptr + 1, 1, 1, stdin) == 1 )

	      /* Peek just one more */

	      *(ptr + 2)= 0 ;
	    else
	      break ;	/* Nothing more to read! */
	  }
	    /* break ; */
	  /* else */
	  if ( SkipCount )		/* Still skipping lines 	*/
	    SkipCount-- ;
	  else
	  if ( SkipToLen ) {            /* Skipping until string found  */
	    if ( strlen(ptr + 1) > SkipToLen &&
		 strncmp(ptr + 1, SkipTo, SkipToLen) == 0 )
	      SkipToLen= 0 ;		/* Gotcha!			*/
	  }
	  else {
	    next= *(ptr + 1) ;		/* Peek char after NewLine	*/
	    if ( *NoJoin && strpos(NoJoin, next) != -1 )
	      next= '\n' ;		/* Flag that we don't join this */
	    if ( next == '\n' ) {       /* Two NewLines=Hard line break */
	      ptr++ ;			/* Skip on past 		*/
	      while ( *ptr++ == '\n' ) ; /* Skip subsequent hard lines	*/
	    }
	    else {                       /* Not a hard NewLine          */
	      if ( (ptr > buff) && (*(ptr - 1) == ' ') ) {

		/* Soft NewLine, but line ended in a space */

		char *target, *source ;

		/* Strip the NewLine out completely */

		target= ptr ;
		source= ptr + 1 ;
		while (*source)
		  *target++ = *source++ ;
		*target= 0 ;
	      }
	      else

		/* Soft NewLine, replace it with a space */

		*ptr= ' ' ;
	    }
	  }
	  ptr++ ;
	}
      }
      fputs(buff, stdout) ;
    }
  }
  return 0 ;
}
